/*
 * pipe.h -- pipeline simulator
 */


#ifndef _PIPE_H_
#define _PIPE_H_


/* MMIX version information */

#define VERSION		1	/* version of the MMIX architecture */
#define SUBVERSION	0	/* secondary byte of version number */
#define SUBSUBVERSION	0	/* further qualification to version number */


/* some renaming to avoid name clashes */

#define random		my_random
#define fsqrt		my_fsqrt
#define div		my_div


/* amount of verbosity */

#define ISSUE_BIT	(1 << 0)   /* show ctrl blocks when issued, etc. */
#define PIPE_BIT	(1 << 1)   /* show the pipeline and locks */
#define COROUTINE_BIT	(1 << 2)   /* show coroutines when started */
#define SCHEDULE_BIT	(1 << 3)   /* show coroutines when scheduled */
#define UNINIT_MEM_BIT	(1 << 4)   /* complain when reading from uninit mem */
#define IO_READ_BIT	(1 << 5)   /* prompt user when reading from I/O loc */
#define SHOW_IO_BIT	(1 << 6)   /* display I/O read/write transactions */
#define SHOW_PRED_BIT	(1 << 7)   /* display branch prediction details */
#define SHOW_CACHE_BIT	(1 << 8)   /* display cache blocks even if invalid */


/* special register numbers */

#define rA		21	/* arithmetic status register */
#define rB		0	/* bootstrap register (trip) */
#define rC		8	/* cycle counter */
#define rD		1	/* dividend register */
#define rE		2	/* epsilon register */
#define rF		22	/* failure location register */
#define rG		19	/* global threshold register */
#define rH		3	/* himult register */
#define rI		12	/* interval counter */
#define rJ		4	/* return-jump register */
#define rK		15	/* interrupt mask register */
#define rL		20	/* local threshold register */
#define rM		5	/* multiplex mask register */
#define rN		9	/* serial number */
#define rO		10	/* register stack offset */
#define rP		23	/* prediction register */
#define rQ		16	/* interrupt request register */
#define rR		6	/* remainder register */
#define rS		11	/* register stack pointer */
#define rT		13	/* trap address register */
#define rU		17	/* usage counter */
#define rV		18	/* virtual translation register */
#define rW		24	/* where-interrupted register (trip) */
#define rX		25	/* execution register (trip) */
#define rY		26	/* Y operand (trip) */
#define rZ		27	/* Z operand (trip) */
#define rBB		7	/* bootstrap register (trap) */
#define rTT		14	/* dynamic trap address register */
#define rWW		28	/* where-interrupted register (trap) */
#define rXX		29	/* execution register (trap) */
#define rYY		30	/* Y operand (trap) */
#define rZZ		31	/* Z operand (trap) */


/* bit codes for trips and traps */

#define P_BIT		(1 << 0)   /* instruction in privileged location */
#define S_BIT		(1 << 1)   /* security violation */
#define B_BIT		(1 << 2)   /* instruction breaks the rules */
#define K_BIT		(1 << 3)   /* instruction for kernel only */
#define N_BIT		(1 << 4)   /* virtual translation bypassed */
#define PX_BIT		(1 << 5)   /* permission lacking to exec from page */
#define PW_BIT		(1 << 6)   /* permission lacking to write on page */
#define PR_BIT		(1 << 7)   /* permission lacking to read from page */
#define PROT_OFFSET	5	   /* distance from PR_BIT to prot code pos */
#define X_BIT		(1 << 8)   /* floating inexact */
#define Z_BIT		(1 << 9)   /* floating division by zero */
#define U_BIT		(1 << 10)  /* floating underflow */
#define O_BIT		(1 << 11)  /* floating overflow */
#define I_BIT		(1 << 12)  /* floating invalid operation */
#define W_BIT		(1 << 13)  /* float-to-fix overflow */
#define V_BIT		(1 << 14)  /* integer overflow */
#define D_BIT		(1 << 15)  /* integer divide check */
#define H_BIT		(1 << 16)  /* trip handler bit */
#define F_BIT		(1 << 17)  /* forced trap bit */
#define E_BIT		(1 << 18)  /* external (dynamic) trap bit */


/* external interrupts */

#define POWER_FAILURE	(1 << 0)   /* try to shut down calmly and quickly */
#define PARITY_ERROR	(1 << 1)   /* try to save the file system */
#define NONACC_MEMORY	(1 << 2)   /* a memory address cannot be used */
#define REBOOT_SIGNAL	(1 << 4)   /* it is time to start over */
#define INTERVAL_TIMER	(1 << 7)   /* the timer register rI has reached 0 */


/* the dispatch stage */

#define UNKNOWN_SPEC	((SpecNode *) 1)

#define X_IS_DEST_BIT	0x20	/* rX is a destination */
#define REL_ADDR_BIT	0x40	/* YZ is part of a relative address */
#define CTL_CHANGE_BIT	0x80	/* control changes at this point */

#define ticks		g[rC].o	/* the internal clock */


/**************************************************************/


typedef struct coroutine {
  char *name;			/* symbolic identification of a coroutine */
  int stage;			/* its rank */
  struct coroutine *next;	/* its successor, but NULL if unscheduled */
  struct coroutine **lockLoc;	/* what it might be locking */
  struct control *ctl;		/* its data */
} Coroutine;

typedef Coroutine *Lockvar;

typedef struct {
  Octa o;			/* the value, if p is NULL */
  struct specNode *p;		/* points to specnode list */
} Spec;

typedef struct specNode {
  Octa o;			/* the value */
  Bool known;			/* is the value already calculated? */
  Octa addr;			/* ?? */
  struct specNode *up, *down;	/* doubly linked specnode list */
} SpecNode;

typedef enum {
  TRAP,   FCMP,   FUN,    FEQL,   FADD,   FIX,    FSUB,   FIXU,
  FLOT,   FLOTI,  FLOTU,  FLOTUI, SFLOT,  SFLOTI, SFLOTU, SFLOTUI,
  FMUL,   FCMPE,  FUNE,   FEQLE,  FDIV,   FSQRT,  FREM,   FINT,
  MUL,    MULI,   MULU,   MULUI,  DIV,    DIVI,   DIVU,   DIVUI,
  ADD,    ADDI,   ADDU,   ADDUI,  SUB,    SUBI,   SUBU,   SUBUI,
  ADDU2,  ADDUI2, ADDU4,  ADDUI4, ADDU8,  ADDUI8, ADDU16, ADDUI16,
  CMP,    CMPI,   CMPU,   CMPUI,  NEG,    NEGI,   NEGU,   NEGUI,
  SL,     SLI,    SLU,    SLUI,   SR,     SRI,    SRU,    SRUI,
  BN,     BNB,    BZ,     BZB,    BP,     BPB,    BOD,    BODB,
  BNN,    BNNB,   BNZ,    BNZB,   BNP,    BNPB,   BEV,    BEVB,
  PBN,    PBNB,   PBZ,    PBZB,   PBP,    PBPB,   PBOD,   PBODB,
  PBNN,   PBNNB,  PBNZ,   PBNZB,  PBNP,   PBNPB,  PBEV,   PBEVB,
  CSN,    CSNI,   CSZ,    CSZI,   CSP,    CSPI,   CSOD,   CSODI,
  CSNN,   CSNNI,  CSNZ,   CSNZI,  CSNP,   CSNPI,  CSEV,   CSEVI,
  ZSN,    ZSNI,   ZSZ,    ZSZI,   ZSP,    ZSPI,   ZSOD,   ZSODI,
  ZSNN,   ZSNNI,  ZSNZ,   ZSNZI,  ZSNP,   ZSNPI,  ZSEV,   ZSEVI,
  LDB,    LDBI,   LDBU,   LDBUI,  LDW,    LDWI,   LDWU,   LDWUI,
  LDT,    LDTI,   LDTU,   LDTUI,  LDO,    LDOI,   LDOU,   LDOUI,
  LDSF,   LDSFI,  LDHT,   LDHTI,  CSWAP,  CSWAPI, LDUNC,  LDUNCI,
  LDVTS,  LDVTSI, PRELD,  PRELDI, PREGO,  PREGOI, GO,     GOI,
  STB,    STBI,   STBU,   STBUI,  STW,    STWI,   STWU,   STWUI,
  STT,    STTI,   STTU,   STTUI,  STO,    STOI,   STOU,   STOUI,
  STSF,   STSFI,  STHT,   STHTI,  STCO,   STCOI,  STUNC,  STUNCI,
  SYNCD,  SYNCDI, PREST,  PRESTI, SYNCID, SYNCIDI,PUSHGO, PUSHGOI,
  OR,     ORI,    ORN,    ORNI,   NOR,    NORI,   XOR,    XORI,
  AND,    ANDI,   ANDN,   ANDNI,  NAND,   NANDI,  NXOR,   NXORI,
  BDIF,   BDIFI,  WDIF,   WDIFI,  TDIF,   TDIFI,  ODIF,   ODIFI,
  MUX,    MUXI,   SADD,   SADDI,  MOR,    MORI,   MXOR,   MXORI,
  SETH,   SETMH,  SETML,  SETL,   INCH,   INCMH,  INCML,  INCL,
  ORH,    ORMH,   ORML,   ORL,    ANDNH,  ANDNMH, ANDNML, ANDNL,
  JMP,    JMPB,   PUSHJ,  PUSHJB, GETA,   GETAB,  PUT,    PUTI,
  POP,    RESUME, SAVE,   UNSAVE, SYNC,   SWYM,   GET,    TRIP
} MmixOpcode;

#define MAX_PIPE_OP	feps	/* any pipeline delays up to here (incl.) */
#define MAX_REAL_CMD	trip	/* internal operations start here (excl.) */

typedef enum {
  mul0,				/* multiplication by zero */
  mul1,				/* multiplication by 1-8 bits */
  mul2,				/* multiplication by 9-16 bits */
  mul3,				/* multiplication by 17-24 bits */
  mul4,				/* multiplication by 25-32 bits */
  mul5,				/* multiplication by 33-40 bits */
  mul6,				/* multiplication by 41-48 bits */
  mul7,				/* multiplication by 49-56 bits */
  mul8,				/* multiplication by 57-64 bits */
  div,				/* DIV[U][I] */
  sh,				/* S[L,R][U][I] */
  mux,				/* MUX[I] */
  sadd,				/* SADD[I] */
  mor,				/* M[X]OR[I] */
  fadd,				/* FADD, FSUB */
  fmul,				/* FMUL */
  fdiv,				/* FDIV */
  fsqrt,			/* FSQRT */
  fint,				/* FINT */
  fix,				/* FIX[U] */
  flot,				/* [S]FLOT[U][I] */
  feps,				/* FCMPE, FUNE, FEQLE */
  fcmp,				/* FCMP */
  funeq,			/* FUN, FEQL */
  fsub,				/* FSUB */
  frem,				/* FREM */
  mul,				/* MUL[I] */
  mulu,				/* MULU[I] */
  divu,				/* DIVU[I] */
  add,				/* ADD[I] */
  addu,				/* [2,4,8,16]ADDU[I], INC[M][H,L]  */
  sub,				/* SUB[I], NEG[I] */
  subu				/* SUBU[I], NEGU[I] */,
  set,				/* SET[M][H,L], GETA[B] */
  or,				/* OR[I], OR[M][H,L] */
  orn,				/* ORN[I] */
  nor,				/* NOR[I] */
  and,				/* AND[I] */
  andn,				/* ANDN[I], ANDN[M][H,L] */
  nand,				/* NAND[I] */
  xor,				/* XOR[I] */
  nxor,				/* NXOR[I] */
  shlu,				/* SLU[I] */
  shru,				/* SRU[I] */
  shl,				/* SL[I] */
  shr,				/* SR[I] */
  cmp,				/* CMP[I] */
  cmpu,				/* CMPU[I] */
  bdif,				/* BDIF[I] */
  wdif,				/* WDIF[I] */
  tdif,				/* TDIF[I] */
  odif,				/* ODIF[I] */
  zset,				/* ZS[N][N,Z,P][I], ZSEV[I], ZSOD[I] */
  cset,				/* CS[N][N,Z,P][I], CSEV[I], CSOD[I] */
  get,				/* GET */
  put,				/* PUT[I] */
  ld,				/* LD[B,W,T,O][U][I], LDHT[I], LDSF[I] */
  ldptp,			/* load page table pointer */
  ldpte,			/* load page table entry */
  ldunc,			/* LDUNC[I] */
  ldvts,			/* LDVTS[I] */
  preld,			/* PRELD[I] */
  prest,			/* PREST[I] */
  st,				/* STO[U][I], STCO[I], STUNC[I] */
  syncd,			/* SYNCD[I] */
  syncid,			/* SYNCID[I] */
  pst,				/* ST[B,W,T][U][I], STHT[I] */
  stunc,			/* STUNC[I], in write buffer */
  cswap,			/* CSWAP[I] */
  br,				/* B[N][N,Z,P][B] */
  pbr,				/* PB[N][N,Z,P][B] */
  pushj,			/* PUSHJ[B] */
  go,				/* GO[I] */
  prego,			/* PREGO[I] */
  pushgo,			/* PUSHGO[I] */
  pop,				/* POP */
  resume,			/* RESUME */
  save,				/* SAVE */
  unsave,			/* UNSAVE */
  sync,				/* SYNC */
  jmp,				/* JMP[B] */
  noop,				/* SWYM */
  trap,				/* TRAP */
  trip,				/* TRIP */
  incgamma,			/* increase gamma pointer */
  decgamma,			/* decrease gamma pointer */
  incrl,			/* increase rL and beta */
  sav,				/* intermediate stage of SAVE */
  unsav,			/* intermediate stage of UNSAVE */
  resum				/* intermediate stage of RESUME */
} InternalOpcode;

typedef struct control {
  Octa loc;			/* virtual addr where the instr originated */
  MmixOpcode op;		/* original instr opcode */
  unsigned char xx, yy, zz;	/* original instr bytes */
  Spec y, z, b, ra;		/* inputs */
  SpecNode x, a, go, rl;	/* outputs */
  Coroutine *owner;		/* a coroutine whose ctl this is */
  InternalOpcode i;		/* internal opcode */
  int state;			/* internal mindset */
  Bool usage;			/* should rU be increased? */
  Bool bNeeded;			/* should we stall until b.p == NULL? */
  Bool raNeeded;		/* should we stall until ra.p == NULL? */
  Bool xIsRen;			/* does x correspond to a rename register? */
  Bool xIsMem;			/* does x correspond to a memory write? */
  Bool aIsRen;			/* does a correspond to a rename register? */
  Bool setL;			/* does rl correspond to a new value of rL? */
  Bool interim;			/* instr needs to be reissued on interrupt? */
  unsigned int arithExc;	/* arith exceptions for event bits of rA */
  unsigned int hist;		/* history bits for branch prediction */
  int denin, denout;		/* exec time penalties for denormals */
  Octa curO, curS;		/* speculative rO and rS before this instr */
  unsigned int interrupt;	/* does this instr generate an interrupt? */
  void *ptrA, *ptrB, *ptrC;	/* generic pointers for miscellaneous use */
} Control;

typedef struct {
  Octa loc;			/* virtual address of instruction */
  Tetra inst;			/* the instruction itself */
  unsigned int interrupt;	/* bit codes that might cause interruption */
  Bool noted;			/* have we peeked at this instruction? */
  unsigned int hist;		/* if we peeked, this was the peekHist */
} Fetch;

typedef struct {
  char name[16];		/* symbolic designation */
  Tetra ops[8];			/* bitmap for the opcodes supported */
  int st;			/* number of pipeline stages */
  Coroutine *co;		/* an array of st coroutines */
} Func;


/**************************************************************/
/* DELETE FROM HERE ------------------------------------------*/
/**************************************************************/


#define PIPE_LIMIT	90

#define WRITE_BACK	1		/* use this if not write-through */
#define WRITE_ALLOC	2		/* use this if not write-around */

#define MAX_STAGE	99
#define VANISH		98
#define FLUSH_TO_MEM	97
#define FLUSH_TO_S	96
#define FILL_FROM_MEM	95
#define FILL_FROM_S	94
#define FILL_FROM_VIRT	93
#define WRITE_FROM_WBUF	92
#define CLEANUP		91


typedef enum {
  random, serial, pseudoLRU, trueLRU
} ReplacePolicy;

typedef struct {
  Octa tag;
  char *dirty;
  Octa *data;
  int rank;
} CacheBlock;

typedef CacheBlock *CacheSet;

typedef struct {
  int a, b, c, g, v;
  int aa, bb, cc, gg, vv;
  int tagMask;
  ReplacePolicy repl, vrepl;
  int mode;
  int accessTime;
  int copyInTime;
  int copyOutTime;
  CacheSet *set;
  CacheSet victim;
  Coroutine filler;
  Control fillerCtl;
  Coroutine flusher;
  Control flusherCtl;
  CacheBlock inbuf;
  CacheBlock outbuf;
  Lockvar lock;
  Lockvar fillLock;
  int ports;
  Coroutine *reader;
  char *name;
} Cache;

typedef struct {
  Tetra tag;
  Octa *chunk;
} ChunkNode;

typedef struct {
  Octa o;		/* data to be stored */
  Octa addr;		/* its physical address */
  Tetra stamp;		/* when last committed (mod 2^32) */
  InternalOpcode i;	/* is this write special? */
} WriteNode;


extern int lastHash;
extern ChunkNode *memHash;
extern Spec instPtr;
extern Octa *fetched;
extern Cache *ITcache;
extern Cache *Icache;
extern Cache *DTcache;
extern Cache *Dcache;
extern Cache *Scache;
extern int pageN;
extern int pageR;
extern int pageS;
extern int pageB[];
extern Octa pageMask;
extern Bool pageBad;
extern Func *funcUnit;
extern int funcUnitCount;
extern int fremMax;
extern int deninPenalty;
extern int denoutPenalty;
extern int holdingTime;
extern int memAddrTime;
extern int busWords;
extern int memReadTime;
extern int memWriteTime;
extern int bpN;
extern int bpA;
extern int bpB;
extern int bpC;
extern char *bpTable;
extern int memChunks;
extern int memChunksMax;
extern int hashPrime;
extern unsigned char pipeSeq[MAX_PIPE_OP + 1][PIPE_LIMIT + 1];
extern WriteNode *wbufBot, *wbufTop;
extern Bool swPT;


void printCache(Cache *cache, Bool invalid);
void printPipe(void);
void printStats(void);
void memWrite(Octa addr, Octa val);
Octa memRead(Octa addr);

int mmgetchars(char *, int, Octa, int);
void mmputchars(unsigned char *, int, Octa);
char stdin_chr(void);


/**************************************************************/
/* DELETE TO HERE --------------------------------------------*/
/**************************************************************/


extern int verbose;		/* controls the level of diagnostic output */

extern int ringSize;		/* size of the ring of queues of coroutines */
extern Coroutine *ring;		/* the ring of queues of coroutines */

extern int fetchMax;		/* max num instrs fetched per clock cycle */
extern int dispatchMax;		/* max num instrs dispatched per clock cycle */
extern int peekAhead;		/* ?? */
extern int commitMax;		/* max num instrs commited per clock cycle */

extern Control *reorderBot;	/* least entry in the reorder buffer ring */
extern Control *reorderTop;	/* greatest entry in the reorder buffer ring */
extern Control *hot;		/* front of the reorder buffer */
extern Control *cool;		/* rear of the reorder buffer */
int deissues;			/* number of instrs that need to be deissued */

extern int *dispatchStat;	/* how often did we dispatch 0,1,... instrs? */
extern Bool noSecurity;		/* omit security checks for testing? */

extern Fetch *fetchBot;		/* least entry in the fetch buffer ring */
extern Fetch *fetchTop;		/* greatest entry in the fetch buffer ring */
extern Fetch *head;		/* front of the fetch buffer */
extern Fetch *tail;		/* rear of the fetch buffer */

extern Func *funit;		/* array of functional units */
extern int funitCount;		/* number of functional units */

extern SpecNode g[256];		/* global and special registers */
extern SpecNode *l;		/* the ring of local registers */
extern int lringSize;		/* number of local registers */
extern int maxRenameRegs;	/* capacity of reorder buffer */
extern int maxMemSlots;		/* capacity of reorder buffer */

extern Octa coolO;		/* value of rO before the cool instruction */
extern Octa coolS;		/* value of rS before the cool instruction */


void printLocks(void);

void mmixInit(void);
void mmixRun(int cycs, Octa breakpoint);


#endif /* _PIPE_H_ */
