/*
 * command.c -- command interpreter
 */


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include "common.h"
#include "console.h"
#include "error.h"
#include "arith.h"
#include "vat.h"
#include "command.h"
#include "spreg.h"
#include "asm.h"
#include "disasm.h"
#include "cpu.h"
#include "memory.h"
#include "timer.h"
#include "term.h"
#include "disk.h"
#include "graph.h"


#define MAX_TOKENS	10


static bool quit;


/**************************************************************/


static char *getOcta(octa virt, octa *valPtr) {
  char *msg;
  octa phys;
  octa data;

  msg = vat(virt, &phys);
  if (msg != NULL) {
    return msg;
  }
  data = memoryRead(phys);
  *valPtr = data;
  return NULL;
}


static char *putOcta(octa virt, octa val) {
  char *msg;
  octa phys;
  octa data;

  msg = vat(virt, &phys);
  if (msg != NULL) {
    return msg;
  }
  data = val;
  memoryWrite(phys, data);
  return NULL;
}


static char *getTetra(octa virt, tetra *valPtr) {
  char *msg;
  octa phys;
  octa data;
  tetra t;

  msg = vat(virt, &phys);
  if (msg != NULL) {
    return msg;
  }
  data = memoryRead(phys);
  switch (phys.l & 0x4) {
    case 0:
      t = data.h;
      break;
    case 4:
      t = data.l;
      break;
  }
  *valPtr = t;
  return NULL;
}


static char *putTetra(octa virt, tetra val) {
  char *msg;
  octa phys;
  octa data;

  msg = vat(virt, &phys);
  if (msg != NULL) {
    return msg;
  }
  data = memoryRead(phys);
  switch (phys.l & 0x4) {
    case 0:
      data.h = val;
      break;
    case 4:
      data.l = val;
      break;
  }
  memoryWrite(phys, data);
  return NULL;
}


static char *getWyde(octa virt, wyde *valPtr) {
  char *msg;
  octa phys;
  octa data;
  wyde w;

  msg = vat(virt, &phys);
  if (msg != NULL) {
    return msg;
  }
  data = memoryRead(phys);
  switch (phys.l & 0x6) {
    case 0:
      w = (data.h >> 16) & 0xFFFF;
      break;
    case 2:
      w = (data.h >>  0) & 0xFFFF;
      break;
    case 4:
      w = (data.l >> 16) & 0xFFFF;
      break;
    case 6:
      w = (data.l >>  0) & 0xFFFF;
      break;
  }
  *valPtr = w;
  return NULL;
}


static char *putWyde(octa virt, wyde val) {
  char *msg;
  octa phys;
  octa data;

  msg = vat(virt, &phys);
  if (msg != NULL) {
    return msg;
  }
  data = memoryRead(phys);
  switch (phys.l & 0x6) {
    case 0:
      data.h &= 0x0000FFFF;
      data.h |= (tetra) val << 16;
      break;
    case 2:
      data.h &= 0xFFFF0000;
      data.h |= (tetra) val <<  0;
      break;
    case 4:
      data.l &= 0x0000FFFF;
      data.l |= (tetra) val << 16;
      break;
    case 6:
      data.l &= 0xFFFF0000;
      data.l |= (tetra) val <<  0;
      break;
  }
  memoryWrite(phys, data);
  return NULL;
}


static char *getByte(octa virt, byte *valPtr) {
  char *msg;
  octa phys;
  octa data;
  byte b;

  msg = vat(virt, &phys);
  if (msg != NULL) {
    return msg;
  }
  data = memoryRead(phys);
  switch (phys.l & 0x7) {
    case 0:
      b = (data.h >> 24) & 0xFF;
      break;
    case 1:
      b = (data.h >> 16) & 0xFF;
      break;
    case 2:
      b = (data.h >>  8) & 0xFF;
      break;
    case 3:
      b = (data.h >>  0) & 0xFF;
      break;
    case 4:
      b = (data.l >> 24) & 0xFF;
      break;
    case 5:
      b = (data.l >> 16) & 0xFF;
      break;
    case 6:
      b = (data.l >>  8) & 0xFF;
      break;
    case 7:
      b = (data.l >>  0) & 0xFF;
      break;
  }
  *valPtr = b;
  return NULL;
}


static char *putByte(octa virt, byte val) {
  char *msg;
  octa phys;
  octa data;

  msg = vat(virt, &phys);
  if (msg != NULL) {
    return msg;
  }
  data = memoryRead(phys);
  switch (phys.l & 0x7) {
    case 0:
      data.h &= 0x00FFFFFF;
      data.h |= (tetra) val << 24;
      break;
    case 1:
      data.h &= 0xFF00FFFF;
      data.h |= (tetra) val << 16;
      break;
    case 2:
      data.h &= 0xFFFF00FF;
      data.h |= (tetra) val <<  8;
      break;
    case 3:
      data.h &= 0xFFFFFF00;
      data.h |= (tetra) val <<  0;
      break;
    case 4:
      data.l &= 0x00FFFFFF;
      data.l |= (tetra) val << 24;
      break;
    case 5:
      data.l &= 0xFF00FFFF;
      data.l |= (tetra) val << 16;
      break;
    case 6:
      data.l &= 0xFFFF00FF;
      data.l |= (tetra) val <<  8;
      break;
    case 7:
      data.l &= 0xFFFFFF00;
      data.l |= (tetra) val <<  0;
      break;
  }
  memoryWrite(phys, data);
  return NULL;
}


/**************************************************************/


static void help(void) {
  cPrintf("valid commands are:\n");
  cPrintf("  +  <num1> <num2>  do some arithmetic on <num1> and <num2>\n");
  cPrintf("  @                 show IP\n");
  cPrintf("  @  <addr>         set IP to <addr>\n");
  cPrintf("  x                 translate IP to physical address\n");
  cPrintf("  x  <addr>         translate <addr> to physical address\n");
  cPrintf("  a                 assemble starting at IP\n");
  cPrintf("  a  <addr>         assemble starting at <addr>\n");
  cPrintf("  u                 unassemble 16 instrs starting at IP\n");
  cPrintf("  u  <addr>         unassemble 16 instrs starting at <addr>\n");
  cPrintf("  u  <addr> <cnt>   unassemble <cnt> instrs starting at <addr>\n");
  cPrintf("  b                 show break\n");
  cPrintf("  b  r              reset break\n");
  cPrintf("  b  <addr>         set break at (virtual) <addr>\n");
  cPrintf("  c                 continue execution\n");
  cPrintf("  c  <cnt>          continue execution <cnt> times\n");
  cPrintf("  s                 single-step one cycle\n");
  cPrintf("  s  <cnt>          single-step <cnt> cycles\n");
  cPrintf("  g                 show available global registers\n");
  cPrintf("  g  <reg>          show global register <reg>\n");
  cPrintf("  g  <reg>  <data>  set global register <reg> to <data>\n");
  cPrintf("  l                 show available local registers\n");
  cPrintf("  l  <reg>          show local register <reg>\n");
  cPrintf("  l  <reg>  <data>  set local register <reg> to <data>\n");
  cPrintf("  $                 show available dynamic registers\n");
  cPrintf("  $  <reg>          show dynamic register <reg>\n");
  cPrintf("  $  <reg>  <data>  set dynamic register <reg> to <data>\n");
  cPrintf("  d  <addr>         dump 256 bytes starting at <addr>\n");
  cPrintf("  d  <addr> <cnt>   dump <cnt> bytes starting at <addr>\n");
  cPrintf("  mo <addr>         show memory octabyte at <addr>\n");
  cPrintf("  mo <addr> <data>  set memory octabyte at <addr> to <data>\n");
  cPrintf("  mt <addr>         show memory tetrabyte at <addr>\n");
  cPrintf("  mt <addr> <data>  set memory tetrabyte at <addr> to <data>\n");
  cPrintf("  mw <addr>         show memory wyde at <addr>\n");
  cPrintf("  mw <addr> <data>  set memory wyde at <addr> to <data>\n");
  cPrintf("  mb <addr>         show memory byte at <addr>\n");
  cPrintf("  mb <addr> <data>  set memory byte at <addr> to <data>\n");
  cPrintf("  IT                show contents of IT cache\n");
  cPrintf("  DT                show contents of DT cache\n");
  cPrintf("  I                 show contents of I cache\n");
  cPrintf("  D                 show contents of D cache\n");
  cPrintf("  D  *              show dirty blocks of D cache\n");
  cPrintf("  S                 show contents of S cache\n");
  cPrintf("  S  *              show dirty blocks of S cache\n");
  cPrintf("  f  s              insert SYNC 5 into the fetch buffer\n");
  cPrintf("  f  <inst>         insert <inst> into the fetch buffer\n");
  cPrintf("  p                 show pipeline and locks\n");
  cPrintf("  v                 show verbose bits\n");
  cPrintf("  v  r              reset verbose bits\n");
  cPrintf("  v  a              set all verbose bits\n");
  cPrintf("  v  <v>            set verbose bits, <v> is any combination ");
  cPrintf("of 'ipcsbw'\n");
  cPrintf("  r                 reset hardware\n");
  cPrintf("  t                 show active timers\n");
  cPrintf("  t  <cnt>          tick time <cnt> times\n");
  cPrintf("  q                 quit simulator\n");
  cPrintf("all numbers may be specified in hexadecimal (#..) ");
  cPrintf("or decimal notation\n");
}


static bool getNum(char *str, octa *val) {
  octa base;
  int digVal;

  if (*str == '#') {
    base.h = 0;
    base.l = 16;
    str++;
    if (!isxdigit(*str)) {
      return false;
    }
  } else {
    base.h = 0;
    base.l = 10;
    if (!isdigit(*str)) {
      return false;
    }
  }
  *val = zero_octa;
  while (isxdigit(*str)) {
    if (*str >= '0' && *str <= '9') {
      digVal = *str - '0';
    } else
    if (*str >= 'A' && *str <= 'F') {
      digVal = *str - 'A' + 10;
    } else
    if (*str >= 'a' && *str <= 'f') {
      digVal = *str - 'a' + 10;
    }
    if (digVal >= base.l) {
      return false;
    }
    *val = incr(omult(*val, base), digVal);
    str++;
  }
  return *str == '\0';
}


static void showIP(void) {
  octa ip;
  tetra instr;
  char *msg;

  cPrintf("IP   ");
  if (cpuTestIP()) {
    ip = cpuGetIP();
    cPrintf("#%08X%08X     ", ip.h, ip.l);
    msg = getTetra(ip, &instr);
    if (msg == NULL) {
      cPrintf("[IP]   #%08X   %s", instr, disasm(instr, ip));
    } else {
      cPrintf("\ncannot read instruction: %s", msg);
    }
  } else {
    cPrintf("--- not known ---");
  }
  cPrintf("\n");
}


static void showBreak(void) {
  octa brk;

  cPrintf("Brk  ");
  if (cpuTestBreak()) {
    brk = cpuGetBreak();
    cPrintf("#%08X%08X", brk.h, brk.l);
  } else {
    cPrintf("---- not set ----");
  }
  cPrintf("\n");
}


static void doArith(char *tokens[], int n) {
  octa num1, num2, res1, res2, res3, res4, res5, res6;

  if (n == 3) {
    if (!getNum(tokens[1], &num1)) {
      cPrintf("illegal first number\n");
      return;
    }
    if (!getNum(tokens[2], &num2)) {
      cPrintf("illegal second number\n");
      return;
    }
    res1 = oplus(num1, num2);
    res2 = ominus(num1, num2);
    res3 = omult(num1, num2);
    res4 = oand(num1, num2);
    res5 = oor(num1, num2);
    res6 = oxor(num1, num2);
    cPrintf("add = #%08X%08X, sub = #%08X%08X, mul = #%08X%08X\n",
            res1, res2, res3);
    cPrintf("and = #%08X%08X, or  = #%08X%08X, xor = #%08X%08X\n",
            res4, res5, res6);
  } else {
    help();
  }
}


static void doIP(char *tokens[], int n) {
  octa addr;

  if (n == 1) {
    showIP();
  } else if (n == 2) {
    if (!getNum(tokens[1], &addr)) {
      cPrintf("illegal address\n");
      return;
    }
    addr.l &= ~0x00000003;
    cpuSetIP(addr);
    showIP();
  } else {
    help();
  }
}


static void doTranslate(char *tokens[], int n) {
  octa virt;
  octa phys;
  char *answer;

  if (n == 1) {
    if (!cpuTestIP()) {
      cPrintf("IP is not known\n");
      return;
    }
    virt = cpuGetIP();
  } else if (n == 2) {
    if (!getNum(tokens[1], &virt)) {
      cPrintf("illegal address\n");
      return;
    }
  } else {
    help();
    return;
  }
  answer = vat(virt, &phys);
  if (answer != NULL) {
    cPrintf("cannot translate #%08X%08X: %s\n",
            virt.h, virt.l, answer);
  } else {
    cPrintf("(virtual) #%08X%08X --> (physical) #%08X%08X\n",
            virt.h, virt.l, phys.h, phys.l);
  }
}


static void doAssemble(char *tokens[], int n) {
  octa addr;
  char prompt[40];
  char *line;
  char *msg;
  tetra instr;

  if (n == 1) {
    if (!cpuTestIP()) {
      cPrintf("IP is not known\n");
      return;
    }
    addr = cpuGetIP();
  } else if (n == 2) {
    if (!getNum(tokens[1], &addr)) {
      cPrintf("illegal address\n");
      return;
    }
  } else {
    help();
    return;
  }
  addr.l &= ~0x00000003;
  while (1) {
    sprintf(prompt, "ASM @ #%08X%08X: ", addr.h, addr.l);
    line = cGetLine(prompt);
    if (*line == '\0' || *line == '\n') {
      break;
    }
    cAddHist(line);
    msg = asmInstr(line, addr, &instr);
    if (msg != NULL) {
      cPrintf("%s\n", msg);
    } else {
      msg = putTetra(addr, instr);
      if (msg != NULL) {
        cPrintf("cannot assemble: %s\n", msg);
        return;
      }
      addr = incr(addr, 4);
    }
  }
}


static void doUnassemble(char *tokens[], int n) {
  octa addr;
  octa count;
  U32 i;
  tetra instr;
  char *msg;

  if (n == 1) {
    if (!cpuTestIP()) {
      cPrintf("IP is not known\n");
      return;
    }
    addr = cpuGetIP();
    count.h = 0;
    count.l = 16;
  } else if (n == 2) {
    if (!getNum(tokens[1], &addr)) {
      cPrintf("illegal address\n");
      return;
    }
    count.h = 0;
    count.l = 16;
  } else if (n == 3) {
    if (!getNum(tokens[1], &addr)) {
      cPrintf("illegal address\n");
      return;
    }
    if (!getNum(tokens[2], &count)) {
      cPrintf("illegal count\n");
      return;
    }
    if (count.h != 0 || count.l > 0x1000) {
      cPrintf("too many lines\n");
      return;
    }
    if (count.h == 0 && count.l == 0) {
      return;
    }
  } else {
    help();
    return;
  }
  addr.l &= ~0x00000003;
  for (i = 0; i < count.l; i++) {
    msg = getTetra(addr, &instr);
    if (msg != NULL) {
      cPrintf("cannot unassemble: %s\n", msg);
      return;
    }
    cPrintf("#%08X%08X:  #%08X    %s\n",
            addr.h, addr.l, instr, disasm(instr, addr));
    addr = incr(addr, 4);
    if (addr.h == 0 && addr.l == 0) {
      /* wrap-around */
      break;
    }
  }
}


static void doBreak(char *tokens[], int n) {
  octa addr;

  if (n == 1) {
    showBreak();
  } else if (n == 2) {
    if (!getNum(tokens[1], &addr)) {
      if (strcmp(tokens[1], "r") != 0) {
        cPrintf("illegal address\n");
        return;
      }
      cpuResetBreak();
      showBreak();
      return;
    }
    addr.l &= ~0x00000003;
    cpuSetBreak(addr);
    showBreak();
  } else {
    help();
  }
}


static void doContinue(char *tokens[], int n) {
  octa count;
  U32 i;
  octa addr;

  if (n == 1) {
    count = one_octa;
  } else if (n == 2) {
    if (!getNum(tokens[1], &count) ||
        count.h != 0 || count.l == 0) {
      cPrintf("illegal count\n");
      return;
    }
  } else {
    help();
    return;
  }
  cPrintf("CPU is running, press ^C to interrupt...\n");
  for (i = 0; i < count.l; i++) {
    cpuRun();
  }
  cPrintf("Break at ");
  if (cpuTestIP()) {
    addr = cpuGetIP();
    cPrintf("#%08X%08X", addr.h, addr.l);
  } else {
    cPrintf("--- not known ---");
  }
  cPrintf("\n");
  showIP();
}


static void doStep(char *tokens[], int n) {
  octa count;
  U32 i;

  if (n == 1) {
    count = one_octa;
  } else if (n == 2) {
    if (!getNum(tokens[1], &count) ||
        count.h != 0 || count.l == 0) {
      cPrintf("illegal count\n");
      return;
    }
  } else {
    help();
    return;
  }
  for (i = 0; i < count.l; i++) {
    cpuStep();
  }
  showIP();
}


static void doGReg(char *tokens[], int n) {
  octa tmp;
  int reg;
  SpReg *rp;
  octa data;
  octa newData;

  if (n == 1) {
    cPrintf("Global registers %d..%d present\n", 0, 255);
    for (reg = 0; reg < 32; reg++) {
      cPrintf("g[%3d]  %-3s  %-27s",
              reg, spRegCodeTbl[reg]->name, spRegCodeTbl[reg]->descr);
      if (reg % 2 == 1) {
        cPrintf("\n");
      }
    }
  } else if (n == 2) {
    if (!getNum(tokens[1], &tmp)) {
      rp = lookupSpReg(tokens[1]);
      if (rp == NULL) {
        cPrintf("illegal register name\n");
        return;
      }
      reg = rp->code;
    } else {
      if (tmp.h != 0 || tmp.l > 255) {
        cPrintf("illegal register number\n");
        return;
      }
      reg = tmp.l;
    }
    data = cpuGetGR(reg);
    if (reg < 32) {
      cPrintf("%-6s  #%08X%08X\n", spRegCodeTbl[reg]->name, data.h, data.l);
    } else {
      cPrintf("g[%3d]  #%08X%08X\n", reg, data.h, data.l);
    }
  } else if (n == 3) {
    if (!getNum(tokens[1], &tmp)) {
      rp = lookupSpReg(tokens[1]);
      if (rp == NULL) {
        cPrintf("illegal register name\n");
        return;
      }
      reg = rp->code;
    } else {
      if (tmp.h != 0 || tmp.l > 255) {
        cPrintf("illegal register number\n");
        return;
      }
      reg = tmp.l;
    }
    if (!getNum(tokens[2], &newData)) {
      cPrintf("illegal data\n");
      return;
    }
    data = cpuGetGR(reg);
    cpuSetGR(reg, newData);
    if (reg < 32) {
      cPrintf("%-6s  #%08X%08X  #%08X%08X\n",
              spRegCodeTbl[reg]->name, data.h, data.l, newData.h, newData.l);
    } else {
      cPrintf("g[%3d]  #%08X%08X  #%08X%08X\n",
              reg, data.h, data.l, newData.h, newData.l);
    }
  } else {
    help();
  }
}


static void doLReg(char *tokens[], int n) {
  octa tmp;
  int reg;
  octa data;
  octa newData;

  if (n == 1) {
    cPrintf("Local registers %d..%d present\n", 0, lring_size - 1);
  } else if (n == 2) {
    if (!getNum(tokens[1], &tmp) || tmp.h != 0 || tmp.l > lring_size - 1) {
      cPrintf("illegal register number\n");
      return;
    }
    reg = tmp.l;
    data = cpuGetLR(reg);
    cPrintf("l[%3d]  #%08X%08X\n", reg, data.h, data.l);
  } else if (n == 3) {
    if (!getNum(tokens[1], &tmp) || tmp.h != 0 || tmp.l > lring_size - 1) {
      cPrintf("illegal register number\n");
      return;
    }
    reg = tmp.l;
    if (!getNum(tokens[2], &newData)) {
      cPrintf("illegal data\n");
      return;
    }
    data = cpuGetLR(reg);
    cpuSetLR(reg, newData);
    cPrintf("l[%3d]  #%08X%08X  #%08X%08X\n",
            reg, data.h, data.l, newData.h, newData.l);
  } else {
    help();
  }
}


static void doDReg(char *tokens[], int n) {
  int L, G;
  octa tmp;
  int reg;
  octa data;
  octa newData;

  L = cpuGetGR(rL).l;
  G = cpuGetGR(rG).l;
  if (n == 1) {
    if (0 <= L - 1) {
      if (0 == L - 1) {
        cPrintf("Dynamic register %d is local\n", 0);
      } else {
        cPrintf("Dynamic registers %d..%d are local\n", 0, L - 1);
      }
    }
    if (L <= G - 1) {
      if (L == G - 1) {
        cPrintf("Dynamic register %s is marginal\n", L);
      } else {
        cPrintf("Dynamic registers %d..%d are marginal\n", L, G - 1);
      }
    }
    if (G <= 255) {
      if (G == 255) {
        cPrintf("Dynamic register %d is global\n", 255);
      } else {
        cPrintf("Dynamic registers %d..%d are global\n", G, 255);
      }
    }
  } else if (n == 2) {
    if (!getNum(tokens[1], &tmp) || tmp.h != 0 || tmp.l > 255) {
      cPrintf("illegal register number\n");
      return;
    }
    reg = tmp.l;
    data = cpuGetDR(reg);
    cPrintf("$[%3d]  #%08X%08X\n", reg, data.h, data.l);
  } else if (n == 3) {
    if (!getNum(tokens[1], &tmp) || tmp.h != 0 || tmp.l > 255) {
      cPrintf("illegal register number\n");
      return;
    }
    if (tmp.l >= L && tmp.l < G) {
      cPrintf("register is marginal and cannot be set\n");
      return;
    }
    reg = tmp.l;
    if (!getNum(tokens[2], &newData)) {
      cPrintf("illegal data\n");
      return;
    }
    data = cpuGetDR(reg);
    cpuSetDR(reg, newData);
    cPrintf("$[%3d]  #%08X%08X  #%08X%08X\n",
            reg, data.h, data.l, newData.h, newData.l);
  } else {
    help();
  }
}


static bool lessThan(octa x, octa y) {
  if (x.h < y.h) {
    return true;
  }
  if (x.h > y.h) {
    return false;
  }
  return x.l < y.l;
}


static bool greaterThan(octa x, octa y) {
  if (x.h > y.h) {
    return true;
  }
  if (x.h < y.h) {
    return false;
  }
  return x.l > y.l;
}


static void doDump(char *tokens[], int n) {
  octa addr;
  octa count;
  octa lo, hi, curr;
  int lines, i, j;
  octa tmp;
  byte c;
  char *msg;

  if (n == 2) {
    if (!getNum(tokens[1], &addr)) {
      cPrintf("illegal address\n");
      return;
    }
    count.h = 0;
    count.l = 16 * 16;
  } else if (n == 3) {
    if (!getNum(tokens[1], &addr)) {
      cPrintf("illegal address\n");
      return;
    }
    if (!getNum(tokens[2], &count)) {
      cPrintf("illegal count\n");
      return;
    }
    if (count.h == 0 && count.l == 0) {
      return;
    }
  } else {
    help();
    return;
  }
  lo.h = addr.h;
  lo.l = addr.l & ~0x0000000F;
  hi = incr(oplus(addr, count), -1);
  if (lessThan(hi, lo)) {
    /* wrap-around */
    hi = neg_one;
  }
  tmp = shift_right(incr(ominus(hi, lo), 16), 4, 1);
  if (tmp.h != 0 || tmp.l > 0x1000) {
    cPrintf("too many lines\n");
    return;
  }
  lines = tmp.l;
  curr = lo;
  cPrintf("#%08X..\n", curr.h);
  for (i = 0; i < lines; i++) {
    cPrintf("..%08X:  ", curr.l);
    for (j = 0; j < 16; j++) {
      tmp = incr(curr, j);
      if (lessThan(tmp, addr) || greaterThan(tmp, hi)) {
        cPrintf("  ");
      } else {
        msg = getByte(tmp, &c);
        if (msg != NULL) {
          cPrintf("\ncannot read memory: %s\n", msg);
          return;
        }
        cPrintf("%02X", c);
      }
      cPrintf(" ");
    }
    cPrintf("  ");
    for (j = 0; j < 16; j++) {
      tmp = incr(curr, j);
      if (lessThan(tmp, addr) || greaterThan(tmp, hi)) {
        cPrintf(" ");
      } else {
        msg = getByte(tmp, &c);
        if (msg != NULL) {
          cPrintf("\ncannot read memory: %s\n", msg);
          return;
        }
        if (c >= 32 && c <= 126) {
          cPrintf("%c", c);
        } else {
          cPrintf(".");
        }
      }
    }
    cPrintf("\n");
    curr = incr(curr, 16);
  }
}


static void doMemoryOcta(char *tokens[], int n) {
  octa addr;
  octa tmpData;
  octa newData;
  octa data;
  char *msg;

  if (n == 2) {
    if (!getNum(tokens[1], &addr)) {
      cPrintf("illegal address\n");
      return;
    }
    addr.l &= ~0x00000007;
    msg = getOcta(addr, &data);
    if (msg != NULL) {
      cPrintf("cannot read octa: %s\n", msg);
      return;
    }
    cPrintf("#%08X%08X:  #%08X%08X\n",
            addr.h, addr.l, data.h, data.l);
  } else if (n == 3) {
    if (!getNum(tokens[1], &addr)) {
      cPrintf("illegal address\n");
      return;
    }
    addr.l &= ~0x00000007;
    if (!getNum(tokens[2], &tmpData)) {
      cPrintf("illegal data\n");
      return;
    }
    newData = tmpData;
    msg = getOcta(addr, &data);
    if (msg != NULL) {
      cPrintf("cannot read octa: %s\n", msg);
      return;
    }
    msg = putOcta(addr, newData);
    if (msg != NULL) {
      cPrintf("cannot write octa: %s\n", msg);
      return;
    }
    cPrintf("#%08X%08X:  #%08X%08X  #%08X%08X\n",
            addr.h, addr.l, data.h, data.l, newData.h, newData.l);
  } else {
    help();
  }
}


static void doMemoryTetra(char *tokens[], int n) {
  octa addr;
  octa tmpData;
  tetra newData;
  tetra data;
  char *msg;

  if (n == 2) {
    if (!getNum(tokens[1], &addr)) {
      cPrintf("illegal address\n");
      return;
    }
    addr.l &= ~0x00000003;
    msg = getTetra(addr, &data);
    if (msg != NULL) {
      cPrintf("cannot read tetra: %s\n", msg);
      return;
    }
    cPrintf("#%08X%08X:  #%08X\n",
            addr.h, addr.l, data);
  } else if (n == 3) {
    if (!getNum(tokens[1], &addr)) {
      cPrintf("illegal address\n");
      return;
    }
    addr.l &= ~0x00000003;
    if (!getNum(tokens[2], &tmpData)) {
      cPrintf("illegal data\n");
      return;
    }
    newData = tmpData.l;
    msg = getTetra(addr, &data);
    if (msg != NULL) {
      cPrintf("cannot read tetra: %s\n", msg);
      return;
    }
    msg = putTetra(addr, newData);
    if (msg != NULL) {
      cPrintf("cannot write tetra: %s\n", msg);
      return;
    }
    cPrintf("#%08X%08X:  #%08X  #%08X\n",
            addr.h, addr.l, data, newData);
  } else {
    help();
  }
}


static void doMemoryWyde(char *tokens[], int n) {
  octa addr;
  octa tmpData;
  wyde newData;
  wyde data;
  char *msg;

  if (n == 2) {
    if (!getNum(tokens[1], &addr)) {
      cPrintf("illegal address\n");
      return;
    }
    addr.l &= ~0x00000001;
    msg = getWyde(addr, &data);
    if (msg != NULL) {
      cPrintf("cannot read wyde: %s\n", msg);
      return;
    }
    cPrintf("#%08X%08X:  #%04X\n",
            addr.h, addr.l, data);
  } else if (n == 3) {
    if (!getNum(tokens[1], &addr)) {
      cPrintf("illegal address\n");
      return;
    }
    addr.l &= ~0x00000001;
    if (!getNum(tokens[2], &tmpData)) {
      cPrintf("illegal data\n");
      return;
    }
    newData = (wyde) tmpData.l;
    msg = getWyde(addr, &data);
    if (msg != NULL) {
      cPrintf("cannot read wyde: %s\n", msg);
      return;
    }
    msg = putWyde(addr, newData);
    if (msg != NULL) {
      cPrintf("cannot write wyde: %s\n", msg);
      return;
    }
    cPrintf("#%08X%08X:  #%04X  #%04X\n",
            addr.h, addr.l, data, newData);
  } else {
    help();
  }
}


static void doMemoryByte(char *tokens[], int n) {
  octa addr;
  octa tmpData;
  byte newData;
  byte data;
  char *msg;

  if (n == 2) {
    if (!getNum(tokens[1], &addr)) {
      cPrintf("illegal address\n");
      return;
    }
    addr.l &= ~0x00000000;
    msg = getByte(addr, &data);
    if (msg != NULL) {
      cPrintf("cannot read byte: %s\n", msg);
      return;
    }
    cPrintf("#%08X%08X:  #%02X\n",
            addr.h, addr.l, data);
  } else if (n == 3) {
    if (!getNum(tokens[1], &addr)) {
      cPrintf("illegal address\n");
      return;
    }
    addr.l &= ~0x00000000;
    if (!getNum(tokens[2], &tmpData)) {
      cPrintf("illegal data\n");
      return;
    }
    newData = (byte) tmpData.l;
    msg = getByte(addr, &data);
    if (msg != NULL) {
      cPrintf("cannot read byte: %s\n", msg);
      return;
    }
    msg = putByte(addr, newData);
    if (msg != NULL) {
      cPrintf("cannot write byte: %s\n", msg);
      return;
    }
    cPrintf("#%08X%08X:  #%02X  #%02X\n",
            addr.h, addr.l, data, newData);
  } else {
    help();
  }
}


static void doIT(char *tokens[], int n) {
  if (n == 1) {
    print_cache(ITcache, false);
  } else {
    help();
  }
}


static void doDT(char *tokens[], int n) {
  if (n == 1) {
    print_cache(DTcache, false);
  } else {
    help();
  }
}


static void doI(char *tokens[], int n) {
  if (n == 1) {
    print_cache(Icache, false);
  } else {
    help();
  }
}


static void doD(char *tokens[], int n) {
  if (n == 1) {
    print_cache(Dcache, false);
  } else if (n == 2) {
    if (strcmp(tokens[1], "*") == 0) {
      print_cache(Dcache, true);
    } else {
      help();
    }
  } else {
    help();
  }
}


static void doS(char *tokens[], int n) {
  if (n == 1) {
    print_cache(Scache, false);
  } else if (n == 2) {
    if (strcmp(tokens[1], "*") == 0) {
      print_cache(Scache, true);
    } else {
      help();
    }
  } else {
    help();
  }
}


static void doFetch(char *tokens[], int n) {
  octa instr;

  if (n == 2) {
    if (!getNum(tokens[1], &instr)) {
      if (strcmp(tokens[1], "s") != 0) {
        cPrintf("illegal instruction\n");
        return;
      }
      instr.h = 0;
      instr.l = 0xFC000005;
    } else {
      if (instr.h != 0) {
        cPrintf("illegal instruction\n");
        return;
      }
    }
  } else {
    help();
    return;
  }
  if (!cpuFetch(instr.l)) {
    cPrintf("cannot fetch: buffer is full\n");
  }
}


static void doPipeline(char *tokens[], int n) {
  if (n == 1) {
    print_pipe();
    print_locks();
  } else {
    help();
  }
}


static void doVerbose(char *tokens[], int n) {
  unsigned int bits;
  char *p;

  if (n == 1) {
    cPrintf("issue (i)        %c\n",
            verbose & issue_bit ? '*' : '-');
    cPrintf("pipe (p)         %c\n",
            verbose & pipe_bit ? '*' : '-');
    cPrintf("coroutine (c)    %c\n",
            verbose & coroutine_bit ? '*' : '-');
    cPrintf("schedule (s)     %c\n",
            verbose & schedule_bit ? '*' : '-');
    cPrintf("branch pred (b)  %c\n",
            verbose & show_pred_bit ? '*' : '-');
    cPrintf("wholecache (w)   %c\n",
            verbose & show_wholecache_bit ? '*' : '-');
  } else if (n == 2) {
    if (strcmp(tokens[1], "r") == 0) {
      verbose = 0;
      return;
    }
    if (strcmp(tokens[1], "a") == 0) {
      verbose = issue_bit | pipe_bit | coroutine_bit |
                schedule_bit | show_pred_bit | show_wholecache_bit;
      return;
    }
    bits = 0;
    if ((p = strchr(tokens[1], 'i')) != NULL) {
      bits |= issue_bit;
      *p = 'U';
    }
    if ((p = strchr(tokens[1], 'p')) != NULL) {
      bits |= pipe_bit;
      *p = 'U';
    }
    if ((p = strchr(tokens[1], 'c')) != NULL) {
      bits |= coroutine_bit;
      *p = 'U';
    }
    if ((p = strchr(tokens[1], 's')) != NULL) {
      bits |= schedule_bit;
      *p = 'U';
    }
    if ((p = strchr(tokens[1], 'b')) != NULL) {
      bits |= show_pred_bit;
      *p = 'U';
    }
    if ((p = strchr(tokens[1], 'w')) != NULL) {
      bits |= show_wholecache_bit;
      *p = 'U';
    }
    p = tokens[1];
    while (*p != '\0') {
      if (*p != 'U') {
        cPrintf("illegal verbose bit specification\n");
        return;
      }
      p++;
    }
    verbose = bits;
  } else {
    help();
  }
}


static void doReset(char *tokens[], int n) {
  if (n == 1) {
    timerReset();
    termReset();
    diskReset();
    graphReset();
    memoryReset();
    cpuReset();
  } else {
    help();
  }
}


static void doTimer(char *tokens[], int n) {
  octa count;
  U32 i;

  if (n == 1) {
    timerShow();
  } else if (n == 2) {
    if (!getNum(tokens[1], &count) ||
        count.h != 0 || count.l == 0) {
      cPrintf("illegal count\n");
      return;
    }
    for (i = 0; i < count.l; i++) {
      timerTick();
    }
  } else {
    help();
  }
}


static void doQuit(char *tokens[], int n) {
  if (n == 1) {
    quit = true;
  } else {
    help();
  }
}


typedef struct {
  char *name;
  void (*cmdProc)(char *tokens[], int n);
} Command;


static Command commands[] = {
  { "+",  doArith },
  { "@",  doIP },
  { "x",  doTranslate },
  { "a",  doAssemble },
  { "u",  doUnassemble },
  { "b",  doBreak },
  { "c",  doContinue },
  { "s",  doStep },
  { "g",  doGReg },
  { "l",  doLReg },
  { "$",  doDReg },
  { "d",  doDump },
  { "mo", doMemoryOcta },
  { "mt", doMemoryTetra },
  { "mw", doMemoryWyde },
  { "mb", doMemoryByte },
  { "IT", doIT },
  { "DT", doDT },
  { "I",  doI },
  { "D",  doD },
  { "S",  doS },
  { "f",  doFetch },
  { "p",  doPipeline },
  { "v",  doVerbose },
  { "r",  doReset },
  { "t",  doTimer },
  { "q",  doQuit }
};


bool execCommand(char *line) {
  char *tokens[MAX_TOKENS];
  int n;
  char *p;
  int i;

  n = 0;
  p = strtok(line, " \t\n");
  while (p != NULL) {
    if (n == MAX_TOKENS) {
      cPrintf("too many tokens on line\n");
      return false;
    }
    tokens[n++] = p;
    p = strtok(NULL, " \t\n");
  }
  if (n == 0) {
    return false;
  }
  for (i = 0; i < sizeof(commands)/sizeof(commands[0]); i++) {
    if (strcmp(commands[i].name, tokens[0]) == 0) {
      quit = false;
      (*commands[i].cmdProc)(tokens, n);
      return quit;
    }
  }
  help();
  return false;
}
