/*
 * disk.c -- disk simulation
 */


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <setjmp.h>

#include "common.h"
#include "console.h"
#include "error.h"
#include "arith.h"
#include "cpu.h"
#include "memory.h"
#include "timer.h"
#include "disk.h"


static bool debug = false;

static FILE *diskImage;
static long numSectors;

static octa diskCtrl;
static octa diskCnt;
static octa diskSct;
static octa diskDma;
static octa diskCap;

static long lastSct;


static octa readOcta(byte *p) {
  octa data;

  data.h = ((tetra) *(p + 0)) << 24 |
           ((tetra) *(p + 1)) << 16 |
           ((tetra) *(p + 2)) <<  8 |
           ((tetra) *(p + 3)) <<  0;
  data.l = ((tetra) *(p + 4)) << 24 |
           ((tetra) *(p + 5)) << 16 |
           ((tetra) *(p + 6)) <<  8 |
           ((tetra) *(p + 7)) <<  0;
  return data;
}


static void writeOcta(byte *p, octa data) {
  *(p + 0) = (byte) (data.h >> 24);
  *(p + 1) = (byte) (data.h >> 16);
  *(p + 2) = (byte) (data.h >>  8);
  *(p + 3) = (byte) (data.h >>  0);
  *(p + 4) = (byte) (data.l >> 24);
  *(p + 5) = (byte) (data.l >> 16);
  *(p + 6) = (byte) (data.l >>  8);
  *(p + 7) = (byte) (data.l >>  0);
}


static void diskCallback(int dev) {
  byte buffer[SECTOR_SIZE];
  int i, j;
  octa data;

  if (debug) {
    cPrintf("\n**** DISK %d CALLBACK ****\n", dev);
  }
  if (diskCap.l != 0 &&
      diskCnt.h == 0 && diskCnt.l != 0 &&
      diskSct.h == 0 && diskSct.l < diskCap.l &&
      diskSct.l + diskCnt.l <= diskCap.l) {
    /* do the transfer */
    if (fseek(diskImage, diskSct.l * SECTOR_SIZE, SEEK_SET) != 0) {
      error("cannot position to sector in disk image");
    }
    if (diskCtrl.l & DISK_WRT) {
      /* memory --> disk */
      for (i = 0; i < diskCnt.l; i++) {
        for (j = 0; j < SECTOR_SIZE; j += 8) {
          data = memoryRead(diskDma);
          writeOcta(buffer + j, data);
          if (debug) {
            cPrintf("\n**** m[#%08X%08X] --> #%08X%08X ****\n",
                    diskDma.h, diskDma.l, data.h, data.l);
          }
          diskDma = incr(diskDma, 8);
        }
        if (fwrite(buffer, SECTOR_SIZE, 1, diskImage) != 1) {
          error("cannot write to disk image");
        }
        diskSct = incr(diskSct, 1);
      }
    } else {
      /* disk --> memory */
      for (i = 0; i < diskCnt.l; i++) {
        if (fread(buffer, SECTOR_SIZE, 1, diskImage) != 1) {
          error("cannot read from disk image");
        }
        for (j = 0; j < SECTOR_SIZE; j += 8) {
          data = readOcta(buffer + j);
          memoryWrite(diskDma, data);
          if (debug) {
            cPrintf("\n**** m[#%08X%08X] <-- #%08X%08X ****\n",
                    diskDma.h, diskDma.l, data.h, data.l);
          }
          diskDma = incr(diskDma, 8);
        }
        diskSct = incr(diskSct, 1);
      }
    }
    lastSct = diskSct.l;
  } else {
    /* sectors requested exceed disk capacity */
    /* or there were zero sectors requested */
    /* or we have no disk at all */
    diskCtrl.l |= DISK_ERR;
  }
  diskCtrl.l &= ~DISK_BUSY;
  if (diskCtrl.l & DISK_IEN) {
    /* raise disk interrupt */
//    cpuSetTrap(IRQ_DISK);
  }
}


octa diskRead(int dev, tetra addr) {
  octa data;

  if (debug) {
    cPrintf("\n**** DISK %d READ from #%08X", dev, addr);
  }
  if (dev >= 1) {
    /* illegal device */
//    throwException(EXC_DT_MEMERR);
  } else
  if (addr == DISK_CTRL) {
    data = diskCtrl;
  } else
  if (addr == DISK_CNT) {
    data = diskCnt;
  } else
  if (addr == DISK_SCT) {
    data = diskSct;
  } else
  if (addr == DISK_DMA) {
    data = diskDma;
  } else
  if (addr == DISK_CAP) {
    data = diskCap;
  } else {
    /* illegal register */
//    throwException(EXC_DT_MEMERR);
  }
  if (debug) {
    cPrintf(", data = #%08X%08X ****\n", data.h, data.l);
  }
  return data;
}


void diskWrite(int dev, tetra addr, octa data) {
  int delay;

  if (debug) {
    cPrintf("\n**** DISK %d WRITE to #%08X, data = #%08X%08X ****\n",
            dev, addr, data.h, data.l);
  }
  if (dev >= 1) {
    /* illegal device */
//    throwException(EXC_DT_MEMERR);
  } else
  if (addr == DISK_CTRL) {
    if (data.l & DISK_WRT) {
      diskCtrl.l |= DISK_WRT;
    } else {
      diskCtrl.l &= ~DISK_WRT;
    }
    if (data.l & DISK_IEN) {
      diskCtrl.l |= DISK_IEN;
    } else {
      diskCtrl.l &= ~DISK_IEN;
    }
    if (data.l & DISK_STRT) {
      diskCtrl.l &= ~DISK_ERR;
      diskCtrl.l |= DISK_BUSY;
      if (diskCap.l != 0) {
        /* disk is present, compute reasonable delay */
        delay = DISK_DELAY +
                (abs(diskSct.l - lastSct) * DISK_SEEK) / diskCap.l;
      } else {
        /* disk is not present, any delay is ok */
        delay = DISK_DELAY;
      }
      timerStart("disk callback", delay, diskCallback, dev);
    }
  } else
  if (addr == DISK_CNT) {
    diskCnt = data;
  } else
  if (addr == DISK_SCT) {
    diskSct = data;
  } else
  if (addr == DISK_DMA) {
    diskDma = data;
  } else
  if (addr == DISK_CAP) {
    /* this register is read-only */
//    throwException(EXC_DT_WRPROT);
  } else {
    /* illegal register */
//    throwException(EXC_DT_MEMERR);
  }
}


void diskReset(void) {
  cPrintf("Resetting Disk...\n");
  diskCtrl.h = 0;
  diskCtrl.l = 0;
  diskCnt.h = 0;
  diskCnt.l = 0;
  diskSct.h = 0;
  diskSct.l = 0;
  diskDma.h = 0;
  diskDma.l = 0;
  diskCap.h = 0;
  diskCap.l = numSectors;
  lastSct = 0;
  if (diskCap.l != 0) {
    cPrintf("Disk of size %ld sectors (%ld bytes) installed.\n",
            numSectors, numSectors * SECTOR_SIZE);
  }
}


void diskInit(char *diskName) {
  long numBytes;

  if (diskName == NULL) {
    /* do not install disk */
    diskImage = NULL;
    numSectors = 0;
  } else {
    /* try to install disk */
    diskImage = fopen(diskName, "r+b");
    if (diskImage == NULL) {
      error("cannot open disk image '%s'", diskName);
    }
    fseek(diskImage, 0, SEEK_END);
    numBytes = ftell(diskImage);
    fseek(diskImage, 0, SEEK_SET);
    if (numBytes % SECTOR_SIZE != 0) {
      error("file '%s' is not a proper disk image", diskName);
    }
    numSectors = numBytes / SECTOR_SIZE;
  }
  diskReset();
}


void diskExit(void) {
  if (diskImage == NULL) {
    /* disk not installed */
    return;
  }
  fclose(diskImage);
}
