/*
 * term.c -- terminal simulation
 */


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <setjmp.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <unistd.h>
#include <fcntl.h>
#include <termios.h>

#include "common.h"
#include "console.h"
#include "error.h"
#include "cpu.h"
#include "timer.h"
#include "term.h"


/**************************************************************/


static bool debug = false;


typedef struct {
  pid_t pid;
  FILE *in;
  FILE *out;
  octa rcvrCtrl;
  octa rcvrData;
  int rcvrIRQ;
  octa xmtrCtrl;
  octa xmtrData;
  int xmtrIRQ;
} Terminal;


static Terminal terminals[NTERMS_MAX];
static int numTerminals;


/**************************************************************/


static void rcvrCallback(int dev) {
  int c;

  if (debug) {
    cPrintf("\n**** TERMINAL %d RCVR CALLBACK ****\n", dev);
  }
  timerStart("terminal rcvr callback", TERM_RCVR_MSEC, rcvrCallback, dev);
  c = fgetc(terminals[dev].in);
  if (c == EOF) {
    /* no character typed */
    return;
  }
  /* any character typed */
  terminals[dev].rcvrData.l = c & 0xFF;
  terminals[dev].rcvrCtrl.l |= TERM_RCVR_RDY;
  if (terminals[dev].rcvrCtrl.l & TERM_RCVR_IEN) {
    /* raise terminal rcvr interrupt */
//    cpuSetTrap(terminals[dev].rcvrIRQ);
  }
}


static void xmtrCallback(int dev) {
  if (debug) {
    cPrintf("\n**** TERMINAL %d XMTR CALLBACK ****\n", dev);
  }
  fputc(terminals[dev].xmtrData.l & 0xFF, terminals[dev].out);
  terminals[dev].xmtrCtrl.l |= TERM_XMTR_RDY;
  if (terminals[dev].xmtrCtrl.l & TERM_XMTR_IEN) {
    /* raise terminal xmtr interrupt */
//    cpuSetTrap(terminals[dev].xmtrIRQ);
  }
}


/**************************************************************/


octa termRead(int dev, tetra addr) {
  octa data;

  if (debug) {
    cPrintf("\n**** TERMINAL %d READ from #%08X", dev, addr);
  }
  if (dev >= numTerminals) {
    /* illegal device */
//    throwException(EXC_DT_MEMERR);
  }
  if (addr == TERM_RCVR_CTRL) {
    data = terminals[dev].rcvrCtrl;
  } else
  if (addr == TERM_RCVR_DATA) {
    terminals[dev].rcvrCtrl.l &= ~TERM_RCVR_RDY;
    data = terminals[dev].rcvrData;
  } else
  if (addr == TERM_XMTR_CTRL) {
    data = terminals[dev].xmtrCtrl;
  } else
  if (addr == TERM_XMTR_DATA) {
    data = terminals[dev].xmtrData;
  } else {
    /* illegal register */
//    throwException(EXC_DT_MEMERR);
  }
  if (debug) {
    cPrintf(", data = #%08X%08X ****\n", data.h, data.l);
  }
  return data;
}


void termWrite(int dev, tetra addr, octa data) {
  if (debug) {
    cPrintf("\n**** TERMINAL %d WRITE to #%08X, data = #%08X%08X ****\n",
            dev, addr, data.h, data.l);
  }
  if (dev >= numTerminals) {
    /* illegal device */
//    throwException(EXC_DT_MEMERR);
  }
  if (addr == TERM_RCVR_CTRL) {
    if (data.l & TERM_RCVR_IEN) {
      terminals[dev].rcvrCtrl.l |= TERM_RCVR_IEN;
    } else {
      terminals[dev].rcvrCtrl.l &= ~TERM_RCVR_IEN;
    }
  } else
  if (addr == TERM_RCVR_DATA) {
    /* this register is read-only */
//    throwException(EXC_DT_WRPROT);
  } else
  if (addr == TERM_XMTR_CTRL) {
    if (data.l & TERM_XMTR_IEN) {
      terminals[dev].xmtrCtrl.l |= TERM_XMTR_IEN;
    } else {
      terminals[dev].xmtrCtrl.l &= ~TERM_XMTR_IEN;
    }
  } else
  if (addr == TERM_XMTR_DATA) {
    terminals[dev].xmtrData.l = data.l & 0xFF;
    terminals[dev].xmtrCtrl.l &= ~TERM_XMTR_RDY;
    timerStart("terminal xmtr callback", TERM_XMTR_MSEC, xmtrCallback, dev);
  } else {
    /* illegal register */
//    throwException(EXC_DT_MEMERR);
  }
}


/**************************************************************/


void termReset(void) {
  int i;

  cPrintf("Resetting Terminals...\n");
  for (i = 0; i < numTerminals; i++) {
    terminals[i].rcvrCtrl.h = 0;
    terminals[i].rcvrCtrl.l = 0;
    terminals[i].rcvrData.h = 0;
    terminals[i].rcvrData.l = 0;
//    terminals[i].rcvrIRQ = IRQ_TERM_0_RCVR + 2 * i;
    timerStart("terminal rcvr callback", TERM_RCVR_MSEC, rcvrCallback, i);
    terminals[i].xmtrCtrl.h = 0;
    terminals[i].xmtrCtrl.l = TERM_XMTR_RDY;
    terminals[i].xmtrData.h = 0;
    terminals[i].xmtrData.l = 0;
//    terminals[i].xmtrIRQ = IRQ_TERM_0_XMTR + 2 * i;
  }
}


static int openPty(int *master, int *slave, char *name) {
  char *p, *q;

  strcpy(name, "/dev/ptyXY");
  for (p = "pqrstuvwxyzPQRST"; *p != '\0'; p++) {
    name[8] = *p;
    for (q = "0123456789abcdef"; *q != '\0'; q++) {
      name[9] = *q;
      /* try to open master */
      *master = open(name, O_RDWR | O_NONBLOCK);
      if (*master < 0) {
        /* open failed, try next one */
        continue;
      }
      /* master opened, try to open slave */
      name[5] = 't';
      *slave = open(name, O_RDWR | O_NONBLOCK);
      if (*slave < 0) {
        /* open failed, close master */
        close(*master);
        return -1;
      }
      /* all is well */
      return 0;
    }
  }
  /* out of pty devices */
  return -1;
}


static void makeRaw(struct termios *tp) {
  tp->c_iflag &= ~(IGNBRK|BRKINT|PARMRK|ISTRIP|INLCR|IGNCR|ICRNL|IXON);
  tp->c_oflag &= ~OPOST;
  tp->c_lflag &= ~(ECHO|ECHONL|ICANON|ISIG|IEXTEN);
  tp->c_cflag &= ~(CSIZE|PARENB);
  tp->c_cflag |= CS8;
}


void termInit(int numTerms) {
  int master, slave;
  char ptyName[50];
  char ptyTitle[50];
  struct termios termios;
  int i;

  numTerminals = numTerms;
  for (i = 0; i < numTerminals; i++) {
    /* open pseudo terminal */
    if (openPty(&master, &slave, ptyName) < 0) {
      error("cannot open pseudo terminal %d", i);
    }
    if (debug) {
      cPrintf("pseudo terminal '%s': master fd = %d, slave fd = %d\n",
              ptyName, master, slave);
    }
    /* set mode to raw */
    tcgetattr(slave, &termios);
    makeRaw(&termios);
    tcsetattr(slave, TCSANOW, &termios);
    /* fork and exec a new xterm */
    terminals[i].pid = fork();
    if (terminals[i].pid < 0) {
      error("cannot fork xterm process %d", i);
    }
    if (terminals[i].pid == 0) {
      /* terminal process */
      setpgid(0, 0);
      close(2);
      close(master);
      sprintf(ptyName, "-Sab%d", slave);
      sprintf(ptyTitle, "GIMMIX Terminal %d", i);
      execlp("xterm", "xterm", "-title", ptyTitle, ptyName, NULL);
      error("cannot exec xterm process %d", i);
    }
    terminals[i].in = fdopen(master, "r");
    setvbuf(terminals[i].in, NULL, _IONBF, 0);
    terminals[i].out = fdopen(master, "w");
    setvbuf(terminals[i].out, NULL, _IONBF, 0);
    /* skip the window id written by xterm */
    while (fgetc(terminals[i].in) != '\n') ;
  }
  termReset();
}


void termExit(void) {
  int i;

  /* kill and wait for all xterm processes */
  for (i = 0; i < numTerminals; i++) {
    kill(terminals[i].pid, SIGKILL);
    waitpid(terminals[i].pid, NULL, 0);
  }
}
