/*
 * showmmo.c -- show the contents of a GIMMIX object file
 */


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <time.h>


/**************************************************************/


#define MAX_SYMLEN	1000		/* max length of a symbol */


#define MM		0x98		/* escape code of mmo format */
#define LOP_QUOTE	0x00		/* quotation lopcode */
#define LOP_LOC		0x01		/* location lopcode */
#define LOP_SKIP	0x02		/* skip lopcode */
#define LOP_FIXO	0x03		/* octabyte-fix lopcode */
#define LOP_FIXR	0x04		/* relative-fix lopcode */
#define LOP_FIXRX	0x05		/* extended relative-fix lopcode */
#define LOP_FILE	0x06		/* file name lopcode */
#define LOP_LINE	0x07		/* file position lopcode */
#define LOP_SPEC	0x08		/* special hook lopcode */
#define LOP_PRE		0x09		/* preamble lopcode */
#define LOP_POST	0x0A		/* postamble lopcode */
#define LOP_STAB	0x0B		/* symbol table lopcode */
#define LOP_END		0x0C		/* end-it-all lopcode */


#define B0(t)		((t >> 24) & 0xFF)
#define B1(t)		((t >> 16) & 0xFF)
#define B2(t)		((t >>  8) & 0xFF)
#define B3(t)		((t >>  0) & 0xFF)

#define W0(t)		((t >> 16) & 0xFFFF)
#define W1(t)		((t >>  0) & 0xFFFF)


/**************************************************************/


typedef unsigned char U8;
typedef signed char S8;
typedef unsigned short U16;
typedef signed short S16;
typedef unsigned int U32;
typedef signed int S32;


typedef U8 byte;
typedef U16 wyde;
typedef U32 tetra;
typedef struct {
  tetra h;
  tetra l;
} octa;


typedef enum { false, true } bool;


/**************************************************************/


bool listing;			/* are we listing everything? */
bool verbose;			/* are we also showing the input? */
bool debug;			/* are we debugging the lopcodes? */
bool xdebug;			/* are we also debugging the symbol table? */

char *mmoName;			/* input file name */
FILE *mmoFile;			/* input file */

octa currLoc;			/* current location */

char *fileName[256];		/* file names seen */
int listedFile;			/* most recently listed file number */
int currFile;			/* most recently selected file number */
int currLine;			/* current position in currFile */

char symBuf[MAX_SYMLEN];	/* the characters up to current node */
char *symPtr;			/* char in symBuf following current prefix */
int byteCount;			/* starts counting at begin of symbol table */
char equBuf[20];		/* equivalent of the current symbol */


/**************************************************************/


void error(char *fmt, ...) {
  va_list ap;

  va_start(ap, fmt);
  fprintf(stderr, "Error: ");
  vfprintf(stderr, fmt, ap);
  fprintf(stderr, "\n");
  va_end(ap);
  exit(1);
}


/**************************************************************/


octa incOcta(octa y, int delta) {
  octa x;

  x.h = y.h;
  x.l = y.l + delta;
  if (delta >= 0) {
    if (x.l < y.l) {
      x.h++;
    }
  } else {
    if (x.l > y.l) {
      x.h--;
    }
  }
  return x;
}


static bool hasUnreadTetra = false;
static tetra unreadValue;


tetra readTetra(void) {
  byte t[4];
  tetra value;

  if (hasUnreadTetra) {
    hasUnreadTetra = false;
    return unreadValue;
  }
  if (fread(t, sizeof(byte), 4, mmoFile) != 4) {
    error("unexpected end of file '%s'", mmoName);
  }
  value = ((tetra) t[0]) << 24 |
          ((tetra) t[1]) << 16 |
          ((tetra) t[2]) <<  8 |
          ((tetra) t[3]) <<  0;
  if (verbose) {
    printf("tetra read: #%08X\n", value);
  }
  return value;
}


void unreadTetra(tetra value) {
  hasUnreadTetra = true;
  unreadValue = value;
}


byte readByte(void) {
  static tetra lastRead;
  byte b;

  switch (byteCount & 0x03) {
    case 0:
      lastRead = readTetra();
      b = B0(lastRead);
      break;
    case 1:
      b = B1(lastRead);
      break;
    case 2:
      b = B2(lastRead);
      break;
    case 3:
      b = B3(lastRead);
      break;
  }
  byteCount++;
  if (xdebug) {
    printf("byte read: #%02X\n", b);
  }
  return b;
}


/**************************************************************/


void usage(char *myself) {
  fprintf(stderr, "Usage: %s [-s] [-v] [-d] [-x] <object file>\n", myself);
  fprintf(stderr, "       -s: list the symbol table only\n");
  fprintf(stderr, "       -v: list tetrabytes of input also\n");
  fprintf(stderr, "       -d: debug loader instructions (implies -v)\n");
  fprintf(stderr, "       -x: extended symbol table debug (implies -d)\n");
  exit(1);
}


void listNormal(tetra value) {
  if (listing) {
    printf("#%08X%08X: #%08X", currLoc.h, currLoc.l, value);
    if (currLine == 0) {
      printf("\n");
    } else {
      if ((currLoc.h & 0xE0000000) != 0) {
        printf("\n");
      } else {
        if (currFile == listedFile) {
          printf(" (line %d)\n", currLine);
        } else {
          printf(" ('%s', line %d)\n", fileName[currFile], currLine);
          listedFile = currFile;
        }
      }
      currLine++;
    }
  }
}


void listStab(void) {
  int ctrl;
  int hiByte, loByte;
  int n;
  int j, k;

  /* read control byte */
  ctrl = readByte();
  if (xdebug) {
    printf("CTRL = #%02X\n", ctrl);
  }
  /* traverse left subtrie if nonempty */
  if (ctrl & 0x40) {
    if (xdebug) {
      printf("left down\n");
    }
    listStab();
    if (xdebug) {
      printf("up, CTRL = #%02X\n", ctrl);
    }
  }
  /* handle current symbol and middle subtrie if present */
  if (ctrl & 0x2F) {
    /* read character */
    if (ctrl & 0x80) {
      /* 16-bit character */
      hiByte = readByte();
    } else {
      /* 8-bit character */
      hiByte = 0;
    }
    loByte = readByte();
    if (hiByte != 0) {
      /* we can't print (hiByte << 8) + loByte without unicode support */
      error("unicode characters in symbols are not supported yet");
    }
    *symPtr++ = loByte;
    if (symPtr == &symBuf[MAX_SYMLEN]) {
      error("symbol too long");
    }
    /* handle current symbol if there is one */
    if (ctrl & 0x0F) {
      *symPtr = '\0';
      n = ctrl & 0x0F;
      if (n == 15) {
        sprintf(equBuf, "$%03d", readByte());
      } else
      if (n <= 8) {
        strcpy(equBuf, "#");
        while (n > 0) {
          sprintf(equBuf + strlen(equBuf), "%02X", readByte());
          n--;
        }
        if (strcmp(equBuf, "#0000") == 0) {
          /* undefined */
          strcpy(equBuf, "?");
        }
      } else {
        strncpy(equBuf, "#20000000000000", 33 - 2 * n);
        equBuf[33 - 2 * n] = '\0';
        while (n > 8) {
          sprintf(equBuf + strlen(equBuf), "%02X", readByte());
          n--;
        }
      }
      k = readByte();
      j = k;
      while (1) {
        if (k >= 128) {
          /* the serial number is now j - 128 */
          break;
        }
        k = readByte();
        j = (j << 7) + k;
      }
      /* print current symbol with its equivalent and serial number */
      printf("    %s = %s (%d)\n", symBuf + 1, equBuf, j - 128);
    }
    /* traverse middle subtrie if nonempty */
    if (ctrl & 0x20) {
      if (xdebug) {
        printf("middle down\n");
      }
      listStab();
      if (xdebug) {
        printf("up, CTRL = #%02X\n", ctrl);
      }
    }
    /* backup */
    symPtr--;
  }
  /* traverse right subtrie if nonempty */
  if (ctrl & 0x10) {
    if (xdebug) {
      printf("right down\n");
    }
    listStab();
    if (xdebug) {
      printf("up, CTRL = #%02X\n", ctrl);
    }
  }
}


int main(int argc, char *argv[]) {
  int i;
  char *argp;
  tetra t;
  bool postambleSeen;
  char *p;
  tetra delta;
  tetra increment;
  octa temp;
  byte b;

  /* process command line */
  listing = true;
  verbose = false;
  debug = false;
  xdebug = false;
  mmoName = NULL;
  for (i = 1; i < argc; i++) {
    argp = argv[i];
    if (*argp == '-') {
      /* option */
      if (strcmp(argp, "-s") == 0) {
        listing = false;
      } else
      if (strcmp(argp, "-v") == 0) {
        verbose = true;
      } else
      if (strcmp(argp, "-d") == 0) {
        verbose = true;
        debug = true;
      } else
      if (strcmp(argp, "-x") == 0) {
        verbose = true;
        debug = true;
        xdebug = true;
      } else {
        usage(argv[0]);
      }
    } else {
      /* file name */
      if (mmoName != NULL) {
        usage(argv[0]);
      }
      mmoName = argp;
    }
  }
  if (mmoName == NULL) {
    usage(argv[0]);
  }
  /* initialize everything */
  mmoFile = fopen(mmoName, "rb");
  if (mmoFile == NULL) {
    error("cannot open object file '%s'", mmoName);
  }
  currLoc.h = 0;
  currLoc.l = 0;
  for (i = 0; i < 256; i++) {
    fileName[i] = NULL;
  }
  listedFile = -1;
  currFile = -1;
  currLine = 0;
  /* list the preamble */
  t = readTetra();
  if (B0(t) != MM || B1(t) != LOP_PRE) {
    error("input file '%s' is not a GIMMIX object file", mmoName);
  }
  if (debug) {
    printf("LOP_PRE (format version %d, %d tetra(s) following)\n",
           B2(t), B3(t));
  }
  if (B2(t) != 1) {
    error("object format version %d not supported", B2(t));
  }
  if (B3(t) > 0) {
    i = B3(t);
    t = readTetra();
    if (listing) {
      printf("File was created %s", asctime(localtime((time_t *) &t)));
    }
    for (i--; i > 0; i--) {
      t = readTetra();
      if (listing) {
        printf("Preamble data #%08X\n", t);
      }
    }
  }
  /* list items */
  postambleSeen = false;
  while (!postambleSeen) {
    t = readTetra();
    if (B0(t) != MM) {
      /* tetra is a normal item */
      listNormal(t);
      currLoc = incOcta(currLoc, 4);
      currLoc.l &= -4;
    } else {
      /* tetra is a loader command */
      switch (B1(t)) {
        case LOP_QUOTE:
          if (debug) {
            printf("LOP_QUOTE (YZ = %d)\n", W1(t));
          }
          if (W1(t) != 1) {
            error("YZ field of lop_quote must be 1");
          }
          t = readTetra();
          listNormal(t);
          currLoc = incOcta(currLoc, 4);
          currLoc.l &= -4;
          break;
        case LOP_LOC:
          if (debug) {
            printf("LOP_LOC (high byte = #%02X, %d tetra(s) following)\n",
                   B2(t), B3(t));
          }
          if (B3(t) == 2) {
            currLoc.h = (B2(t) << 24) + readTetra();
          } else
          if (B3(t) == 1) {
            currLoc.h = B2(t) << 24;
          } else {
            error("Z field of lop_loc must be 1 or 2");
          }
          currLoc.l = readTetra();
          if (debug) {
            printf("Set location to #%08X%08X\n", currLoc.h, currLoc.l);
          }
          break;
        case LOP_SKIP:
          if (debug) {
            printf("LOP_SKIP (delta = #%04X)\n", W1(t));
          }
          delta = W1(t);
          currLoc = incOcta(currLoc, delta);
          if (debug) {
            printf("Skip to #%08X%08X\n", currLoc.h, currLoc.l);
          }
          break;
        case LOP_FIXO:
          if (debug) {
            printf("LOP_FIXO (high byte = #%02X, %d tetra(s) following)\n",
                   B2(t), B3(t));
          }
          if (B3(t) == 2) {
            temp.h = (B2(t) << 24) + readTetra();
          } else
          if (B3(t) == 1) {
            temp.h = B2(t) << 24;
          } else {
            error("Z field of lop_fixo must be 1 or 2");
          }
          temp.l = readTetra();
          if (listing) {
            printf("#%08X%08X: #%08X%08X\n",
                   temp.h, temp.l, currLoc.h, currLoc.l);
          }
          break;
        case LOP_FIXR:
          if (debug) {
            printf("LOP_FIXR (delta = #%04X)\n", W1(t));
          }
          delta = W1(t);
          increment = -delta << 2;
          temp = incOcta(currLoc, increment);
          if (listing) {
            printf("#%08X%08X: #%08X\n", temp.h, temp.l, delta);
          }
          break;
        case LOP_FIXRX:
          if (debug) {
            printf("LOP_FIXRX (Y = %d, Z = %d)\n", B2(t), B3(t));
          }
          if (B2(t) != 0) {
            error("Y field of lop_fixrx must be 0");
          }
          if (B3(t) != 16 && B3(t) != 24) {
            error("Z field of lop_fixrx must be 16 or 24");
          }
          delta = readTetra();
          if ((delta & 0xFE000000) != 0) {
            error("increment of lop_fixrx is too large");
          }
          if (delta >= 0x1000000) {
            increment = -((delta & 0xFFFFFF) - (1 << B3(t))) << 2;
          } else {
            increment = -delta << 2;
          }
          temp = incOcta(currLoc, increment);
          if (listing) {
            printf("#%08X%08X: #%08X\n", temp.h, temp.l, delta);
          }
          break;
        case LOP_FILE:
          if (debug) {
            printf("LOP_FILE (file number = %d, %d tetra(s) following)\n",
                   B2(t), B3(t));
          }
          if (fileName[B2(t)] != NULL) {
            if (B3(t) != 0) {
              error("two file names with the same number");
            }
            currFile = B2(t);
            if (debug) {
              printf("Switch to file number %d\n", currFile);
            }
          } else {
            if (B3(t) == 0) {
              error("no name given for newly selected file");
            }
            fileName[B2(t)] = calloc(4 * B3(t) + 1, 1);
            if (fileName[B2(t)] == NULL) {
              error("no room to store the file name");
            }
            currFile = B2(t);
            p = fileName[currFile];
            for (i = B3(t); i > 0; i--) {
              t = readTetra();
              *(p + 0) = B0(t);
              *(p + 1) = B1(t);
              *(p + 2) = B2(t);
              *(p + 3) = B3(t);
              p += 4;
            }
            if (debug) {
              printf("Switch to file number %d with name '%s'\n",
                     currFile, fileName[currFile]);
            }
          }
          currLine = 0;
          break;
        case LOP_LINE:
          if (debug) {
            printf("LOP_LINE (line number %d)\n", W1(t));
          }
          if (currFile < 0) {
            error("no file was selected for lop_line");
          }
          currLine = W1(t);
          if (debug) {
            printf("Set line to %d\n", currLine);
          }
          break;
        case LOP_SPEC:
          if (debug) {
            printf("LOP_SPEC (type = %d)\n", W1(t));
          }
          if (listing) {
            printf("Special data %d at loc #%08X%08X",
                   W1(t), currLoc.h, currLoc.l);
            if (currLine == 0) {
              printf("\n");
            } else
            if (currFile == listedFile) {
              printf(" (line %d)\n", currLine);
            } else {
              printf(" ('%s', line %d)\n", fileName[currFile], currLine);
              listedFile = currFile;
            }
          }
          while (1) {
            t = readTetra();
            if (B0(t) == MM) {
              if (B1(t) != LOP_QUOTE || W1(t) != 1) {
                /* end of special data */
                unreadTetra(t);
                break;
              }
              t = readTetra();
            }
            if (listing) {
              printf("                    #%08X\n", t);
            }
          }
          break;
        case LOP_PRE:
          if (debug) {
            printf("LOP_PRE\n");
          }
          error("cannot have another preamble");
          break;
        case LOP_POST:
          if (debug) {
            printf("LOP_POST (Y = %d, Z = %d)\n", B2(t), B3(t));
          }
          postambleSeen = true;
          if (B2(t) != 0) {
            error("Y field of lop_post must be zero");
          }
          if (B3(t) < 32) {
            error("Z field of lop_post is less than 32");
          }
          break;
        case LOP_STAB:
          if (debug) {
            printf("LOP_STAB\n");
          }
          error("symbol table must follow postamble");
          break;
        case LOP_END:
          if (debug) {
            printf("LOP_END\n");
          }
          error("symbol table cannot end before it begins");
          break;
        default:
          if (debug) {
            printf("LOP_???\n");
          }
          error("unknown loader command #%02X", B1(t));
          break;
      }
    }
  }
  /* list the postamble */
  for (i = B3(t); i < 256; i++) {
    temp.h = readTetra();
    temp.l = readTetra();
    if (listing) {
      if (temp.h != 0 || temp.l != 0) {
        printf("g%03d: #%08X%08X\n", i, temp.h, temp.l);
      } else {
        printf("g%03d: 0\n", i);
      }
    }
  }
  /* list the symbol table */
  t = readTetra();
  if (B0(t) != MM || B1(t) != LOP_STAB) {
    error("no symbol table following the postamble");
  }
  if (debug) {
    printf("LOP_STAB\n");
  }
  if (W1(t) != 0) {
    error("YZ field of lop_stab must be zero");
  }
  printf("Symbol table (beginning at tetra %ld):\n",
         ftell(mmoFile) / sizeof(tetra));
  symPtr = symBuf;
  byteCount = 0;
  listStab();
  while ((byteCount & 0x03) != 0) {
    if (readByte() != 0) {
      error("nonzero byte follows the symbol table");
    }
  }
  t = readTetra();
  if (B0(t) != MM || B1(t) != LOP_END) {
    error("symbol table is not followed by lop_end");
  }
  if (debug) {
    printf("LOP_END (YZ = %d)\n", W1(t));
  }
  if (W1(t) != byteCount / sizeof(tetra)) {
    error("YZ field of lop_end is %d, should have been %d",
          W1(t), byteCount / sizeof(tetra));
  } else {
    if (verbose) {
      printf("Symbol table ends at tetra %ld\n",
             ftell(mmoFile) / sizeof(tetra));
    }
    if (fread(&b, 1, 1, mmoFile) != 0) {
      error("extra bytes follow the lop_end");
    }
  }
  /* close object file and return */
  fclose(mmoFile);
  return 0;
}
